@@ -22,10 +22,13 @@
 #include <gtk/css/gtkcss.h>
 #include "gtk/css/gtkcsstokenizerprivate.h"
 #include "gtk/css/gtkcssparserprivate.h"
+#include "gtk/css/gtkcssvariablevalueprivate.h"
 #include "gtkbitmaskprivate.h"
 #include "gtkcssarrayvalueprivate.h"
 #include "gtkcsscolorvalueprivate.h"
+#include "gtkcsscustompropertypoolprivate.h"
 #include "gtkcsskeyframesprivate.h"
+#include "gtkcssreferencevalueprivate.h"
 #include "gtkcssselectorprivate.h"
 #include "gtkcssshorthandpropertyprivate.h"
 #include "gtksettingsprivate.h"
@@ -49,6 +52,9 @@
 #define GDK_ARRAY_PREALLOC 64
 #include "gdk/gdkarrayimpl.c"
 
+/* For lack of a better place, assert here that these two definitions match */
+G_STATIC_ASSERT (GTK_DEBUG_CSS == GTK_CSS_PARSER_DEBUG_CSS);
+
 /**
  * GtkCssProvider:
  *
@@ -113,6 +119,7 @@ struct GtkCssRuleset
   PropertyValue *styles;
   guint n_styles;
   guint owns_styles : 1;
+  GHashTable *custom_properties;
 };
 
 struct _GtkCssScanner
@@ -133,6 +140,7 @@ struct _GtkCssProviderPrivate
   GtkCssSelectorTree *tree;
   GResource *resource;
   char *path;
+  GBytes *bytes; /* *no* reference */
 };
 
 enum {
@@ -179,10 +187,12 @@ gtk_css_provider_parsing_error (GtkCssProvider  *provider,
     {
       char *s = gtk_css_section_to_string (section);
 
-      g_warning ("Theme parser %s: %s: %s",
-                 error->domain == GTK_CSS_PARSER_WARNING ? "warning" : "error",
-                 s,
-                 error->message);
+      if (GTK_DEBUG_CHECK (CSS) ||
+          !g_error_matches (error, GTK_CSS_PARSER_WARNING, GTK_CSS_PARSER_WARNING_DEPRECATED))
+        g_warning ("Theme parser %s: %s: %s",
+                   error->domain == GTK_CSS_PARSER_WARNING ? "warning" : "error",
+                   s,
+                   error->message);
 
       g_free (s);
     }
@@ -213,6 +223,9 @@ gtk_css_provider_class_init (GtkCssProviderClass *klass)
    *
    * Signals that a parsing error occurred.
    *
+   * The expected error values are in the [error@Gtk.CssParserError]
+   * and [enum@Gtk.CssParserWarning] enumerations.
+   *
    * The @path, @line and @position describe the actual location of
    * the error as accurately as possible.
    *
@@ -221,6 +234,9 @@ gtk_css_provider_class_init (GtkCssProviderClass *klass)
    * even all of it to not be parsed at all. So it is a useful idea
    * to check that the parsing succeeds by connecting to this signal.
    *
+   * Errors in the [enum@Gtk.CssParserWarning] enumeration should not
+   * be treated as fatal errors.
+   *
    * Note that this signal may be emitted at any time as the css provider
    * may opt to defer parsing parts or all of the input to a later time
    * than when a loading function was called.
@@ -264,12 +280,14 @@ gtk_css_ruleset_clear (GtkCssRuleset *ruleset)
 
       for (i = 0; i < ruleset->n_styles; i++)
         {
-          _gtk_css_value_unref (ruleset->styles[i].value);
+          gtk_css_value_unref (ruleset->styles[i].value);
 	  ruleset->styles[i].value = NULL;
 	  if (ruleset->styles[i].section)
 	    gtk_css_section_unref (ruleset->styles[i].section);
         }
       g_free (ruleset->styles);
+      if (ruleset->custom_properties)
+        g_hash_table_unref (ruleset->custom_properties);
     }
   if (ruleset->selector)
     _gtk_css_selector_free (ruleset->selector);
@@ -285,7 +303,7 @@ gtk_css_ruleset_add (GtkCssRuleset       *ruleset,
 {
   guint i;
 
-  g_return_if_fail (ruleset->owns_styles || ruleset->n_styles == 0);
+  g_return_if_fail (ruleset->owns_styles || (ruleset->n_styles == 0 && ruleset->custom_properties == NULL));
 
   ruleset->owns_styles = TRUE;
 
@@ -293,7 +311,7 @@ gtk_css_ruleset_add (GtkCssRuleset       *ruleset,
     {
       if (ruleset->styles[i].property == property)
         {
-          _gtk_css_value_unref (ruleset->styles[i].value);
+          gtk_css_value_unref (ruleset->styles[i].value);
 	  ruleset->styles[i].value = NULL;
 	  if (ruleset->styles[i].section)
 	    gtk_css_section_unref (ruleset->styles[i].section);
@@ -315,6 +333,39 @@ gtk_css_ruleset_add (GtkCssRuleset       *ruleset,
     ruleset->styles[i].section = NULL;
 }
 
+static void
+unref_custom_property_name (gpointer pointer)
+{
+  GtkCssCustomPropertyPool *pool = gtk_css_custom_property_pool_get ();
+
+  gtk_css_custom_property_pool_unref (pool, GPOINTER_TO_INT (pointer));
+}
+
+static void
+gtk_css_ruleset_add_custom (GtkCssRuleset       *ruleset,
+                            const char          *name,
+                            GtkCssVariableValue *value)
+{
+  GtkCssCustomPropertyPool *pool;
+  int id;
+
+  g_return_if_fail (ruleset->owns_styles || (ruleset->n_styles == 0 && ruleset->custom_properties == NULL));
+
+  ruleset->owns_styles = TRUE;
+
+  if (ruleset->custom_properties == NULL)
+    {
+      ruleset->custom_properties = g_hash_table_new_full (g_direct_hash, g_direct_equal,
+                                                          unref_custom_property_name,
+                                                          (GDestroyNotify) gtk_css_variable_value_unref);
+    }
+
+  pool = gtk_css_custom_property_pool_get ();
+  id = gtk_css_custom_property_pool_add (pool, name);
+
+  g_hash_table_replace (ruleset->custom_properties, GINT_TO_POINTER (id), value);
+}
+
 static void
 gtk_css_scanner_destroy (GtkCssScanner *scanner)
 {
@@ -342,9 +393,10 @@ gtk_css_scanner_parser_error (GtkCssParser         *parser,
   GtkCssScanner *scanner = user_data;
   GtkCssSection *section;
 
-  section = gtk_css_section_new (gtk_css_parser_get_file (parser),
-                                 start,
-                                 end);
+  section = gtk_css_section_new_with_bytes (gtk_css_parser_get_file (parser),
+                                            gtk_css_parser_get_bytes (parser),
+                                            start,
+                                            end);
 
   gtk_css_style_provider_emit_error (GTK_STYLE_PROVIDER (scanner->provider), section, error);
 
@@ -399,7 +451,7 @@ gtk_css_provider_init (GtkCssProvider *css_provider)
 
   priv->symbolic_colors = g_hash_table_new_full (g_str_hash, g_str_equal,
                                                  (GDestroyNotify) g_free,
-                                                 (GDestroyNotify) _gtk_css_value_unref);
+                                                 (GDestroyNotify) gtk_css_value_unref);
   priv->keyframes = g_hash_table_new_full (g_str_hash, g_str_equal,
                                            (GDestroyNotify) g_free,
                                            (GDestroyNotify) _gtk_css_keyframes_unref);
@@ -490,7 +542,7 @@ gtk_css_style_provider_lookup (GtkStyleProvider             *provider,
         {
           ruleset = gtk_css_selector_matches_get (&tree_rules, i);
 
-          if (ruleset->styles == NULL)
+          if (ruleset->styles == NULL && ruleset->custom_properties == NULL)
             continue;
 
           for (j = 0; j < ruleset->n_styles; j++)
@@ -506,6 +558,18 @@ gtk_css_style_provider_lookup (GtkStyleProvider             *provider,
                                    ruleset->styles[j].section,
                                    ruleset->styles[j].value);
             }
+
+          if (ruleset->custom_properties)
+            {
+              GHashTableIter iter;
+              gpointer id;
+              GtkCssVariableValue *value;
+
+              g_hash_table_iter_init (&iter, ruleset->custom_properties);
+
+              while (g_hash_table_iter_next (&iter, &id, (gpointer) &value))
+                _gtk_css_lookup_set_custom (lookup, GPOINTER_TO_INT (id), value);
+            }
         }
     }
   gtk_css_selector_matches_clear (&tree_rules);
@@ -514,6 +578,16 @@ gtk_css_style_provider_lookup (GtkStyleProvider             *provider,
     *change = gtk_css_selector_tree_get_change_all (priv->tree, filter, node);
 }
 
+static gboolean
+gtk_css_style_provider_has_section (GtkStyleProvider *provider,
+                                    GtkCssSection    *section)
+{
+  GtkCssProvider *self = GTK_CSS_PROVIDER (provider);
+  GtkCssProviderPrivate *priv = gtk_css_provider_get_instance_private (self);
+
+  return priv->bytes == gtk_css_section_get_bytes (section);
+}
+
 static void
 gtk_css_style_provider_iface_init (GtkStyleProviderInterface *iface)
 {
@@ -521,6 +595,7 @@ gtk_css_style_provider_iface_init (GtkStyleProviderInterface *iface)
   iface->get_keyframes = gtk_css_style_provider_get_keyframes;
   iface->lookup = gtk_css_style_provider_lookup;
   iface->emit_error = gtk_css_style_provider_emit_error;
+  iface->has_section = gtk_css_style_provider_has_section;
 }
 
 static void
@@ -572,7 +647,7 @@ css_provider_commit (GtkCssProvider  *css_provider,
   GtkCssProviderPrivate *priv = gtk_css_provider_get_instance_private (css_provider);
   guint i;
 
-  if (ruleset->styles == NULL)
+  if (ruleset->styles == NULL && ruleset->custom_properties == NULL)
     {
       for (i = 0; i < gtk_css_selectors_get_size (selectors); i++)
         _gtk_css_selector_free (gtk_css_selectors_get (selectors, i));
@@ -704,7 +779,7 @@ parse_color_definition (GtkCssScanner *scanner)
   if (name == NULL)
     return TRUE;
 
-  color = _gtk_css_color_value_parse (scanner->parser);
+  color = gtk_css_color_value_parse (scanner->parser);
   if (color == NULL)
     {
       g_free (name);
@@ -714,7 +789,7 @@ parse_color_definition (GtkCssScanner *scanner)
   if (!gtk_css_parser_has_token (scanner->parser, GTK_CSS_TOKEN_EOF))
     {
       g_free (name);
-      _gtk_css_value_unref (color);
+      gtk_css_value_unref (color);
       gtk_css_parser_error_syntax (scanner->parser,
                                    "Missing semicolon at end of color definition");
       return TRUE;
@@ -814,6 +889,49 @@ parse_declaration (GtkCssScanner *scanner,
   if (name == NULL)
     goto out;
 
+  /* This is a custom property */
+  if (name[0] == '-' && name[1] == '-')
+    {
+      GtkCssVariableValue *value;
+      GtkCssLocation start_location;
+      GtkCssSection *section;
+
+      if (!gtk_css_parser_try_token (scanner->parser, GTK_CSS_TOKEN_COLON))
+        {
+          gtk_css_parser_error_syntax (scanner->parser, "Expected ':'");
+          goto out;
+        }
+
+      gtk_css_parser_skip_whitespace (scanner->parser);
+
+      if (gtk_keep_css_sections)
+        start_location = *gtk_css_parser_get_start_location (scanner->parser);
+
+      value = gtk_css_parser_parse_value_into_token_stream (scanner->parser);
+      if (value == NULL)
+        goto out;
+
+      if (gtk_keep_css_sections)
+        {
+          section = gtk_css_section_new_with_bytes (gtk_css_parser_get_file (scanner->parser),
+                                                    gtk_css_parser_get_bytes (scanner->parser),
+                                                    &start_location,
+                                                    gtk_css_parser_get_start_location (scanner->parser));
+        }
+      else
+        section = NULL;
+
+      if (section != NULL)
+        {
+          gtk_css_variable_value_set_section (value, section);
+          gtk_css_section_unref (section);
+        }
+
+      gtk_css_ruleset_add_custom (ruleset, name, value);
+
+      goto out;
+    }
+
   property = _gtk_style_property_lookup (name);
 
   if (property)
@@ -827,22 +945,88 @@ parse_declaration (GtkCssScanner *scanner,
           goto out;
         }
 
-      value = _gtk_style_property_parse_value (property, scanner->parser);
-
-      if (value == NULL)
-        goto out;
-
-      if (!gtk_css_parser_has_token (scanner->parser, GTK_CSS_TOKEN_EOF))
+      if (gtk_css_parser_has_references (scanner->parser))
         {
-          gtk_css_parser_error_syntax (scanner->parser, "Junk at end of value for %s", property->name);
-          goto out;
+          GtkCssLocation start_location;
+          GtkCssVariableValue *var_value;
+
+          gtk_css_parser_skip_whitespace (scanner->parser);
+
+          if (gtk_keep_css_sections)
+            start_location = *gtk_css_parser_get_start_location (scanner->parser);
+
+          var_value = gtk_css_parser_parse_value_into_token_stream (scanner->parser);
+          if (var_value == NULL)
+            goto out;
+
+          if (gtk_keep_css_sections)
+            section = gtk_css_section_new_with_bytes (gtk_css_parser_get_file (scanner->parser),
+                                                      gtk_css_parser_get_bytes (scanner->parser),
+                                                      &start_location,
+                                                      gtk_css_parser_get_start_location (scanner->parser));
+          else
+            section = NULL;
+
+          if (section != NULL)
+            {
+              gtk_css_variable_value_set_section (var_value, section);
+              gtk_css_section_unref (section);
+            }
+
+          if (GTK_IS_CSS_SHORTHAND_PROPERTY (property))
+            {
+              GtkCssShorthandProperty *shorthand = GTK_CSS_SHORTHAND_PROPERTY (property);
+              guint i, n;
+              GtkCssValue **values;
+
+              n = _gtk_css_shorthand_property_get_n_subproperties (shorthand);
+
+              values = g_new (GtkCssValue *, n);
+
+              for (i = 0; i < n; i++)
+                {
+                  GtkCssValue *child =
+                    _gtk_css_reference_value_new (property,
+                                                  var_value,
+                                                  gtk_css_parser_get_file (scanner->parser));
+                  _gtk_css_reference_value_set_subproperty (child, i);
+
+                  values[i] = _gtk_css_array_value_get_nth (child, i);
+                }
+
+              value = _gtk_css_array_value_new_from_array (values, n);
+              g_free (values);
+            }
+          else
+            {
+              value = _gtk_css_reference_value_new (property,
+                                                    var_value,
+                                                    gtk_css_parser_get_file (scanner->parser));
+            }
+
+          gtk_css_variable_value_unref (var_value);
+        }
+      else
+        {
+          value = _gtk_style_property_parse_value (property, scanner->parser);
+
+          if (value == NULL)
+            goto out;
+
+          if (!gtk_css_parser_has_token (scanner->parser, GTK_CSS_TOKEN_EOF))
+            {
+              gtk_css_parser_error_syntax (scanner->parser, "Junk at end of value for %s", property->name);
+              gtk_css_value_unref (value);
+              goto out;
+            }
         }
 
       if (gtk_keep_css_sections)
         {
-          section = gtk_css_section_new (gtk_css_parser_get_file (scanner->parser),
-                                         gtk_css_parser_get_block_location (scanner->parser),
-                                         gtk_css_parser_get_end_location (scanner->parser));
+          section = gtk_css_section_new_with_bytes (gtk_css_parser_get_file (scanner->parser),
+                                                    gtk_css_parser_get_bytes (scanner->parser),
+                                                    gtk_css_parser_get_block_location (scanner->parser),
+                                                    gtk_css_parser_get_end_location (scanner->parser));
         }
       else
         section = NULL;
@@ -857,10 +1041,10 @@ parse_declaration (GtkCssScanner *scanner,
               GtkCssStyleProperty *child = _gtk_css_shorthand_property_get_subproperty (shorthand, i);
               GtkCssValue *sub = _gtk_css_array_value_get_nth (value, i);
 
-              gtk_css_ruleset_add (ruleset, child, _gtk_css_value_ref (sub), section);
+              gtk_css_ruleset_add (ruleset, child, gtk_css_value_ref (sub), section);
             }
 
-            _gtk_css_value_unref (value);
+            gtk_css_value_unref (value);
         }
       else if (GTK_IS_CSS_STYLE_PROPERTY (property))
         {
@@ -870,7 +1054,7 @@ parse_declaration (GtkCssScanner *scanner,
       else
         {
           g_assert_not_reached ();
-          _gtk_css_value_unref (value);
+          gtk_css_value_unref (value);
         }
 
       g_clear_pointer (&section, gtk_css_section_unref);
@@ -1025,6 +1209,7 @@ gtk_css_provider_load_internal (GtkCssProvider *self,
                                 GFile          *file,
                                 GBytes         *bytes)
 {
+  GtkCssProviderPrivate *priv = gtk_css_provider_get_instance_private (self);
   gint64 before G_GNUC_UNUSED;
 
   before = GDK_PROFILER_CURRENT_TIME;
@@ -1059,6 +1244,8 @@ gtk_css_provider_load_internal (GtkCssProvider *self,
         }
     }
 
+  priv->bytes = bytes;
+
   if (bytes)
     {
       GtkCssScanner *scanner;
@@ -1112,7 +1299,7 @@ gtk_css_provider_load_from_data (GtkCssProvider  *css_provider,
   if (length < 0)
     length = strlen (data);
 
-  bytes = g_bytes_new_static (data, length);
+  bytes = g_bytes_new (data, length);
 
   gtk_css_provider_load_from_bytes (css_provider, bytes);
 
@@ -1139,7 +1326,7 @@ gtk_css_provider_load_from_string (GtkCssProvider *css_provider,
   g_return_if_fail (GTK_IS_CSS_PROVIDER (css_provider));
   g_return_if_fail (string != NULL);
 
-  bytes = g_bytes_new_static (string, strlen (string));
+  bytes = g_bytes_new (string, strlen (string));
 
   gtk_css_provider_load_from_bytes (css_provider, bytes);
 
@@ -1231,25 +1418,18 @@ gtk_css_provider_load_from_path (GtkCssProvider  *css_provider,
  */
 void
 gtk_css_provider_load_from_resource (GtkCssProvider *css_provider,
-			             const char     *resource_path)
+                                     const char     *resource_path)
 {
-  GFile *file;
-  char *uri, *escaped;
+  GBytes *bytes;
 
   g_return_if_fail (GTK_IS_CSS_PROVIDER (css_provider));
   g_return_if_fail (resource_path != NULL);
 
-  escaped = g_uri_escape_string (resource_path,
-				 G_URI_RESERVED_CHARS_ALLOWED_IN_PATH, FALSE);
-  uri = g_strconcat ("resource://", escaped, NULL);
-  g_free (escaped);
+  bytes = g_resources_lookup_data (resource_path, G_RESOURCE_LOOKUP_FLAGS_NONE, NULL);
 
-  file = g_file_new_for_uri (uri);
-  g_free (uri);
+  gtk_css_provider_load_from_bytes (css_provider, bytes);
 
-  gtk_css_provider_load_from_file (css_provider, file);
-
-  g_object_unref (file);
+  g_bytes_unref (bytes);
 }
 
 char *
@@ -1282,39 +1462,38 @@ _gtk_css_provider_get_theme_dir (GtkCssProvider *provider)
 
 /*
  * Look for
- * $dir/$subdir/gtk-4.16/gtk-$variant.css
- * $dir/$subdir/gtk-4.14/gtk-$variant.css
+ * $dir/$subdir/gtk-4.16/$file
+ * $dir/$subdir/gtk-4.14/$file
  *  ...
- * $dir/$subdir/gtk-4.0/gtk-$variant.css
+ * $dir/$subdir/gtk-4.0/$file
  * and return the first found file.
  */
 static char *
 _gtk_css_find_theme_dir (const char *dir,
                          const char *subdir,
                          const char *name,
-                         const char *variant)
+                         const char *file)
 {
-  char *file;
-  char *base;
-  char *subsubdir;
-  int i;
   char *path;
-
-  if (variant)
-    file = g_strconcat ("gtk-", variant, ".css", NULL);
-  else
-    file = g_strdup ("gtk.css");
+  char *base;
 
   if (subdir)
     base = g_build_filename (dir, subdir, name, NULL);
   else
     base = g_build_filename (dir, name, NULL);
 
-  for (i = MINOR; i >= 0; i = i - 2)
+  if (!g_file_test (base, G_FILE_TEST_IS_DIR))
     {
-      subsubdir = g_strdup_printf ("gtk-4.%d", i);
+      g_free (base);
+      return NULL;
+    }
+
+  for (int i = MINOR; i >= 0; i = i - 2)
+    {
+      char subsubdir[64];
+
+      g_snprintf (subsubdir, sizeof (subsubdir), "gtk-4.%d", i);
       path = g_build_filename (base, subsubdir, file, NULL);
-      g_free (subsubdir);
 
       if (g_file_test (path, G_FILE_TEST_EXISTS))
         break;
@@ -1323,7 +1502,6 @@ _gtk_css_find_theme_dir (const char *dir,
       path = NULL;
     }
 
-  g_free (file);
   g_free (base);
 
   return path;
@@ -1335,18 +1513,24 @@ static char *
 _gtk_css_find_theme (const char *name,
                      const char *variant)
 {
+  char file[256];
   char *path;
   const char *const *dirs;
   int i;
   char *dir;
 
+  if (variant)
+    g_snprintf (file, sizeof (file), "gtk-%s.css", variant);
+  else
+    strcpy (file, "gtk.css");
+
   /* First look in the user's data directory */
-  path = _gtk_css_find_theme_dir (g_get_user_data_dir (), "themes", name, variant);
+  path = _gtk_css_find_theme_dir (g_get_user_data_dir (), "themes", name, file);
   if (path)
     return path;
 
   /* Next look in the user's home directory */
-  path = _gtk_css_find_theme_dir (g_get_home_dir (), ".themes", name, variant);
+  path = _gtk_css_find_theme_dir (g_get_home_dir (), ".themes", name, file);
   if (path)
     return path;
 
@@ -1354,14 +1538,14 @@ _gtk_css_find_theme (const char *name,
   dirs = g_get_system_data_dirs ();
   for (i = 0; dirs[i]; i++)
     {
-      path = _gtk_css_find_theme_dir (dirs[i], "themes", name, variant);
+      path = _gtk_css_find_theme_dir (dirs[i], "themes", name, file);
       if (path)
         return path;
     }
 
   /* Finally, try in the default theme directory */
   dir = _gtk_get_theme_dir ();
-  path = _gtk_css_find_theme_dir (dir, NULL, name, variant);
+  path = _gtk_css_find_theme_dir (dir, NULL, name, file);
   g_free (dir);
 
   return path;
@@ -1448,7 +1632,7 @@ gtk_css_provider_load_named (GtkCssProvider *provider,
           else
             gtk_css_provider_load_named (provider, DEFAULT_THEME_NAME, "hc");
         }
-      else if (strcmp (name, "HighConstrastInverse") == 0)
+      else if (strcmp (name, "HighContrastInverse") == 0)
         gtk_css_provider_load_named (provider, DEFAULT_THEME_NAME, "hc-dark");
       else if (strcmp (name, "Adwaita-dark") == 0)
         gtk_css_provider_load_named (provider, DEFAULT_THEME_NAME, "dark");
@@ -1473,6 +1657,22 @@ compare_properties (gconstpointer a, gconstpointer b, gpointer style)
                  _gtk_style_property_get_name (GTK_STYLE_PROPERTY (styles[*ub].property)));
 }
 
+/* This is looking into a GPtrArray where each "pointer" is actually
+ * GINT_TO_POINTER (id), so a and b are pointers to pointer-sized quantities */
+static int
+compare_custom_properties (gconstpointer a, gconstpointer b, gpointer user_data)
+{
+  GtkCssCustomPropertyPool *pool = user_data;
+  const void * const *ap = a;
+  const void * const *bp = b;
+  const char *name1, *name2;
+
+  name1 = gtk_css_custom_property_pool_get_name (pool, GPOINTER_TO_INT (*ap));
+  name2 = gtk_css_custom_property_pool_get_name (pool, GPOINTER_TO_INT (*bp));
+
+  return strcmp (name1, name2);
+}
+
 static void
 gtk_css_ruleset_print (const GtkCssRuleset *ruleset,
                        GString             *str)
@@ -1499,13 +1699,38 @@ gtk_css_ruleset_print (const GtkCssRuleset *ruleset,
           g_string_append (str, "  ");
           g_string_append (str, _gtk_style_property_get_name (GTK_STYLE_PROPERTY (prop->property)));
           g_string_append (str, ": ");
-          _gtk_css_value_print (prop->value, str);
+          gtk_css_value_print (prop->value, str);
           g_string_append (str, ";\n");
         }
 
       g_free (sorted);
     }
 
+  if (ruleset->custom_properties)
+    {
+      GtkCssCustomPropertyPool *pool = gtk_css_custom_property_pool_get ();
+      GPtrArray *keys;
+
+      keys = g_hash_table_get_keys_as_ptr_array (ruleset->custom_properties);
+      g_ptr_array_sort_with_data (keys, compare_custom_properties, pool);
+
+      for (i = 0; i < keys->len; i++)
+        {
+          int id = GPOINTER_TO_INT (g_ptr_array_index (keys, i));
+          const char *name = gtk_css_custom_property_pool_get_name (pool, id);
+          GtkCssVariableValue *value = g_hash_table_lookup (ruleset->custom_properties,
+                                                            GINT_TO_POINTER (id));
+
+          g_string_append (str, "  ");
+          g_string_append (str, name);
+          g_string_append (str, ": ");
+          gtk_css_variable_value_print (value, str);
+          g_string_append (str, ";\n");
+        }
+
+      g_ptr_array_unref (keys);
+    }
+
   g_string_append (str, "}\n");
 }
 
@@ -1527,7 +1752,7 @@ gtk_css_provider_print_colors (GHashTable *colors,
       g_string_append (str, "@define-color ");
       g_string_append (str, name);
       g_string_append (str, " ");
-      _gtk_css_value_print (color, str);
+      gtk_css_value_print (color, str);
       g_string_append (str, ";\n");
     }
 
@@ -1598,4 +1823,3 @@ gtk_css_provider_to_string (GtkCssProvider *provider)
 
   return g_string_free (str, FALSE);
 }
-
